/*

Copyright 2012 Adam Reichold

This file is part of qpdfview.

qpdfview is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

qpdfview is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with qpdfview.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "document.h"
#include "documentview.h"
#include "miscellaneous.h"

// auxiliary view

AuxiliaryView::AuxiliaryView(QWidget *parent) : QWidget(parent),
    m_mainView(0)
{
}

DocumentView *AuxiliaryView::mainView() const
{
    return m_mainView;
}

void AuxiliaryView::setMainView(DocumentView *mainView)
{
    if(m_mainView != mainView)
    {
        m_mainView = mainView;

        if(m_mainView)
        {
            connect(m_mainView->m_document, SIGNAL(documentChanged()), this, SLOT(slotDocumentChanged()));
        }

        if(this->isVisible())
        {
            this->slotDocumentChanged();
        }
    }
}

void AuxiliaryView::showEvent(QShowEvent *event)
{
    if(!event->spontaneous())
    {
        this->slotDocumentChanged();
    }
}

// outline view

OutlineView::OutlineView(QWidget *parent) : AuxiliaryView(parent)
{
    m_treeWidget = new QTreeWidget(this);

    m_treeWidget->setAlternatingRowColors(true);
    m_treeWidget->header()->setVisible(false);

    this->setLayout(new QVBoxLayout());
    this->layout()->setContentsMargins(0, 0, 0, 0);
    this->layout()->addWidget(m_treeWidget);

    connect(m_treeWidget, SIGNAL(itemClicked(QTreeWidgetItem*,int)), this, SLOT(slotItemClicked(QTreeWidgetItem*,int)));
}

void OutlineView::slotDocumentChanged()
{
    m_treeWidget->clear();

    if(mainView())
    {
        TocNode *node = mainView()->m_document->toc();

        if(node)
        {
            prepareOutline(node, 0, 0);

            delete node;
        }

    }
}

void OutlineView::slotItemClicked(QTreeWidgetItem *item, int column)
{
    int page = item->data(column, Qt::UserRole).toInt();
    qreal top = item->data(column, Qt::UserRole+1).toReal();

    mainView()->setCurrentPage(page, top);
}

void OutlineView::prepareOutline(TocNode *node, QTreeWidgetItem *parent, QTreeWidgetItem *sibling)
{
    QTreeWidgetItem *item = 0;

    if(parent)
    {
        item = new QTreeWidgetItem(parent);
    }
    else
    {
        item = new QTreeWidgetItem(m_treeWidget, sibling);
    }

    item->setText(0, node->text);
    item->setData(0, Qt::UserRole, node->page);
    item->setData(0, Qt::UserRole+1, node->top);

    if(node->child)
    {
        prepareOutline(node->child, item, 0);
    }

    if(node->sibling)
    {
        prepareOutline(node->sibling, parent, item);
    }
}

// thumbnails view

ThumbnailsView::PageItem::PageItem(Document *document, int index, QGraphicsItem *parent, QGraphicsScene *scene) : QGraphicsItem(parent, scene),
    m_document(document),
    m_index(index),
    m_size(),
    m_render()
{
    m_size = m_document->size(m_index);
}

ThumbnailsView::PageItem::~PageItem()
{
    if(m_render.isRunning())
    {
        m_render.waitForFinished();
    }
}

QRectF ThumbnailsView::PageItem::boundingRect() const
{
    return QRectF(0.0, 0.0, qCeil(0.1 * m_size.width()), qCeil(0.1 * m_size.height()));
}

void ThumbnailsView::PageItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *)
{
    // draw page

    painter->fillRect(boundingRect(), QBrush(Qt::white));

    QImage image = m_document->pullPage(m_index, 0.1);

    if(image.isNull())
    {
        if(!m_render.isRunning())
        {
            m_render = QtConcurrent::run(this, &PageItem::render);
        }
    }
    else
    {
        painter->drawImage(QPointF(), image);
    }

    painter->setPen(QPen(Qt::black));
    painter->drawRect(boundingRect());
}

void ThumbnailsView::PageItem::mousePressEvent(QGraphicsSceneMouseEvent*)
{
    ThumbnailsView *parent = qobject_cast<ThumbnailsView*>(scene()->parent());

    if(!parent)
    {
        qFatal("!parent");
        return;
    }

    parent->mainView()->setCurrentPage(m_index + 1);
}

void ThumbnailsView::PageItem::render()
{
    bool visible = false;

    QRectF pageRect = boundingRect().translated(pos());

    foreach(QGraphicsView *view, scene()->views())
    {
       QRectF viewRect = view->mapToScene(view->rect()).boundingRect();

       visible = visible || viewRect.intersects(pageRect);
    }

    if(visible)
    {
       m_document->pushPage(m_index, 0.1);
    }

    scene()->update(boundingRect().translated(pos()));
}

ThumbnailsView::ThumbnailsView(QWidget *parent) : AuxiliaryView(parent)
{
    m_scene = new QGraphicsScene(this);
    m_scene->setBackgroundBrush(QBrush(Qt::darkGray));

    m_view = new QGraphicsView(m_scene, this);
    m_view->show();
}

void ThumbnailsView::resizeEvent(QResizeEvent *event)
{
    m_view->resize(event->size());
}

void ThumbnailsView::slotDocumentChanged()
{
    m_scene->clear();

    if(mainView())
    {
        qreal sceneWidth = 0.0, sceneHeight = 5.0;

        for(int index = 0; index < mainView()->m_document->numberOfPages(); index++)
        {
            PageItem *pageItem = new PageItem(mainView()->m_document, index);

            pageItem->setPos(5.0, sceneHeight);

            sceneWidth = qMax(sceneWidth, pageItem->boundingRect().width() + 10.0);
            sceneHeight += pageItem->boundingRect().height() + 5.0;

            m_scene->addItem(pageItem);

            QGraphicsSimpleTextItem *textItem = new QGraphicsSimpleTextItem(QLocale::system().toString(index + 1));

            textItem->setPos(10.0, sceneHeight);

            m_scene->addItem(textItem);

            sceneWidth = qMax(sceneWidth, textItem->boundingRect().width() + 15.0);
            sceneHeight += textItem->boundingRect().height() + 5.0;
        }

        m_scene->setSceneRect(0.0, 0.0, sceneWidth, sceneHeight);
        m_view->setSceneRect(0.0, 0.0, sceneWidth, sceneHeight);

        this->setMinimumWidth(static_cast<int>(sceneWidth) + 35);
    }
}

// presentation view

PresentationView::PresentationView() : AuxiliaryView(),
    m_currentPage(1),
    m_size(),
    m_scaleFactor(1.0),
    m_boundingRect(),
    m_links(),
    m_linkTransform(),
    m_render()
{
    Qt::WindowFlags flags = this->windowFlags();
    flags = flags | Qt::FramelessWindowHint;
    this->setWindowFlags(flags);

    Qt::WindowStates states = this->windowState();
    states = states | Qt::WindowFullScreen;
    this->setWindowState(states);

    QPalette palette = this->palette();
    palette.setColor(QPalette::Window, Qt::black);
    this->setPalette(palette);

    this->setMouseTracking(true);
}

PresentationView::~PresentationView()
{
    if(m_render.isRunning())
    {
        m_render.waitForFinished();
    }
}

void PresentationView::setCurrentPage(int currentPage)
{
    if(m_currentPage != currentPage && currentPage >= 1 && currentPage <= mainView()->m_document->numberOfPages())
    {
        m_currentPage = currentPage;

        this->prepareView();
    }
}

void PresentationView::previousPage()
{
    if(m_currentPage > 1)
    {
        m_currentPage--;

        this->prepareView();
    }
}

void PresentationView::nextPage()
{
    if(m_currentPage < mainView()->m_document->numberOfPages())
    {
        m_currentPage++;

        this->prepareView();
    }
}

void PresentationView::firstPage()
{
    if(m_currentPage != 1)
    {
        m_currentPage = 1;

        this->prepareView();
    }
}

void PresentationView::lastPage()
{
    if(m_currentPage != mainView()->m_document->numberOfPages())
    {
        m_currentPage = mainView()->m_document->numberOfPages();

        this->prepareView();
    }
}

void PresentationView::paintEvent(QPaintEvent *event)
{
    QWidget::paintEvent(event);

    QPainter painter(this);

    painter.fillRect(m_boundingRect, QBrush(Qt::white));

    QImage image = mainView()->m_document->pullPage(m_currentPage-1, m_scaleFactor);

    if(image.isNull())
    {
        if(!m_render.isRunning())
        {
            m_render = QtConcurrent::run(this, &PresentationView::render);
        }
    }
    else
    {
        painter.drawImage(m_boundingRect.topLeft(), image);
    }

    painter.setPen(QPen(Qt::black));
    painter.drawRect(m_boundingRect);
}

void PresentationView::resizeEvent(QResizeEvent*)
{
    this->prepareView();
}

void PresentationView::keyPressEvent(QKeyEvent *event)
{
    switch(event->key())
    {
    case Qt::Key_PageUp:
    case Qt::Key_Up:
    case Qt::Key_Left:
    case Qt::Key_Backspace:
        this->previousPage();

        break;
    case Qt::Key_PageDown:
    case Qt::Key_Down:
    case Qt::Key_Right:
    case Qt::Key_Space:
        this->nextPage();

        break;
    case Qt::Key_Home:
        this->firstPage();

        break;
    case Qt::Key_End:
        this->lastPage();

        break;
    case Qt::Key_F12:
    case Qt::Key_Escape:
        this->close();

        break;
    default:
        QWidget::keyPressEvent(event);

        break;
    }
}

void PresentationView::mousePressEvent(QMouseEvent *event)
{
    foreach(Link link, m_links)
    {
        if(m_linkTransform.mapRect(link.area).contains(event->posF()))
        {
            if(!link.url.isEmpty())
            {
                if(Document::openUrl())
                {
                    QDesktopServices::openUrl(QUrl(link.url));
                }
            }
            else
            {
                this->setCurrentPage(link.page);
            }

            return;
        }
    }
}

void PresentationView::mouseMoveEvent(QMouseEvent *event)
{
    QApplication::restoreOverrideCursor();

    foreach(Link link, m_links)
    {
        if(m_linkTransform.mapRect(link.area).contains(event->posF()))
        {
            QApplication::setOverrideCursor(Qt::PointingHandCursor);

            if(!link.url.isEmpty())
            {
                QToolTip::showText(event->globalPos(), tr("Open URL \"%1\".").arg(link.url));
            }
            else
            {
                QToolTip::showText(event->globalPos(), tr("Go to page %1.").arg(link.page));
            }

            return;
        }
    }

    QToolTip::hideText();
}

void PresentationView::slotDocumentChanged()
{
    m_currentPage = mainView()->currentPage();

    this->prepareView();
}

void PresentationView::prepareView()
{
    m_size = mainView()->m_document->size(m_currentPage-1);

    m_scaleFactor = qMin(static_cast<qreal>(this->width()) / m_size.width(), static_cast<qreal>(this->height()) / m_size.height());

    m_boundingRect.setLeft(0.5 * (static_cast<qreal>(this->width()) - m_scaleFactor * m_size.width()));
    m_boundingRect.setTop(0.5 * (static_cast<qreal>(this->height()) - m_scaleFactor * m_size.height()));
    m_boundingRect.setWidth(m_scaleFactor * m_size.width());
    m_boundingRect.setHeight(m_scaleFactor * m_size.height());

    m_links = mainView()->m_document->links(m_currentPage-1);
    m_linkTransform = QTransform(m_scaleFactor * m_size.width(), 0.0, 0.0, m_scaleFactor * m_size.height(), m_boundingRect.left(), m_boundingRect.top());

    this->update();
}

void PresentationView::render()
{
    mainView()->m_document->pushPage(m_currentPage-1, m_scaleFactor);

    this->update();
}

// recently used action

RecentlyUsedAction::RecentlyUsedAction(QObject *parent) : QAction(tr("Recently &used"), parent),
    m_settings()
{
    m_menu = new QMenu();
    this->setMenu(m_menu);

    m_actionGroup = new QActionGroup(this);
    connect(m_actionGroup, SIGNAL(triggered(QAction*)), this, SLOT(slotActionGroupTriggered(QAction*)));

    m_clearListAction = new QAction(tr("&Clear list"), this);
    connect(m_clearListAction, SIGNAL(triggered()), this, SLOT(clearList()));

    m_separator = m_menu->addSeparator();
    m_menu->addAction(m_clearListAction);

    QStringList filePaths = m_settings.value("mainWindow/recentlyUsed").toStringList();

    foreach(QString filePath, filePaths)
    {
        QAction *action = new QAction(this);
        action->setText(filePath);
        action->setData(filePath);

        m_actionGroup->addAction(action);
        m_menu->insertAction(m_separator, action);
    }
}

RecentlyUsedAction::~RecentlyUsedAction()
{
    QStringList filePaths;

    foreach(QAction *action, m_actionGroup->actions())
    {
        filePaths.append(action->data().toString());
    }

    m_settings.setValue("mainWindow/recentlyUsed", filePaths);

    delete m_menu;
}

void RecentlyUsedAction::addEntry(const QString &filePath)
{
    bool addItem = true;

    foreach(QAction *action, m_actionGroup->actions())
    {
        addItem = addItem && action->data().toString() != filePath;
    }

    if(addItem)
    {
        QAction *action = new QAction(this);
        action->setText(filePath);
        action->setData(filePath);

        m_actionGroup->addAction(action);
        m_menu->insertAction(m_separator, action);
    }

    if(m_actionGroup->actions().size() > 5)
    {
        QAction *first = m_actionGroup->actions().first();

        m_actionGroup->removeAction(first);
        m_menu->removeAction(first);
    }
}

void RecentlyUsedAction::clearList()
{
    foreach(QAction *action, m_actionGroup->actions())
    {
        m_actionGroup->removeAction(action);
        m_menu->removeAction(action);
    }
}

void RecentlyUsedAction::slotActionGroupTriggered(QAction *action)
{
    emit entrySelected(action->data().toString());
}

// bookmarks menu

BookmarksMenu::BookmarksMenu(QWidget *parent) : QMenu(tr("Bookmarks"), parent),
    m_pages(),
    m_currentPage(-1),
    m_value(-1)
{
    m_actionGroup = new QActionGroup(this);
    connect(m_actionGroup, SIGNAL(triggered(QAction*)), this, SLOT(slotActionGroupTriggered(QAction*)));

    m_addEntryAction = new QAction(tr("&Add entry"), this);
    m_addEntryAction->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_B));
    connect(m_addEntryAction, SIGNAL(triggered()), this, SLOT(addEntry()));

    m_removeEntriesOnCurrentPageAction = new QAction(tr("&Remove entries on current page"), this);
    m_removeEntriesOnCurrentPageAction->setShortcut(QKeySequence(Qt::CTRL + Qt::SHIFT + Qt::Key_B));
    connect(m_removeEntriesOnCurrentPageAction, SIGNAL(triggered()), this, SLOT(removeEntriesOnCurrentPage()));

    m_goToPreviousEntryAction = new QAction(tr("Go to &previous entry"), this);
    m_goToPreviousEntryAction->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_PageUp));
    connect(m_goToPreviousEntryAction, SIGNAL(triggered()), this, SLOT(goToPreviousEntry()));

    m_goToNextEntryAction = new QAction(tr("Go to &next entry"), this);
    m_goToNextEntryAction->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_PageDown));
    connect(m_goToNextEntryAction, SIGNAL(triggered()), this, SLOT(goToNextEntry()));

    m_clearListAction = new QAction(tr("&Clear list"), this);
    m_clearListAction->setShortcut(QKeySequence(Qt::CTRL + Qt::ALT + Qt::Key_B));
    connect(m_clearListAction, SIGNAL(triggered()), this, SLOT(clearList()));

    this->addAction(m_addEntryAction);
    this->addAction(m_removeEntriesOnCurrentPageAction);
    this->addAction(m_goToPreviousEntryAction);
    this->addAction(m_goToNextEntryAction);
    this->addSeparator();
    this->addAction(m_clearListAction);
    this->addSeparator();
}

void BookmarksMenu::addEntry()
{
    bool addItem = true;
    QAction *before = 0;

    foreach(QAction *action, m_actionGroup->actions())
    {
        if(QPair<int, int>(m_pages.value(action), action->data().toInt()) == QPair<int, int>(m_currentPage, m_value))
        {
            addItem = false;

            break;
        }
        else if(QPair<int, int>(m_pages.value(action), action->data().toInt()) > QPair<int,int>(m_currentPage, m_value))
        {
            if(before)
            {
                before = QPair<int, int>(m_pages.value(action), action->data().toInt()) < QPair<int, int>(m_pages.value(before), before->data().toInt()) ? action : before;
            }
            else
            {
                before = action;
            }
        }
    }

    if(addItem)
    {
        QAction *action = new QAction(this);
        action->setText(tr("%2 on page %1").arg(m_currentPage).arg(m_value));
        action->setData(m_value);

        m_pages.insert(action, m_currentPage);

        m_actionGroup->addAction(action);
        this->insertAction(before, action);
    }

    if(m_actionGroup->actions().size() > 10)
    {
        QAction *first = m_actionGroup->actions().first();

        m_actionGroup->removeAction(first);
        this->removeAction(first);

        m_pages.remove(first);;
        first->deleteLater();
    }
}

void BookmarksMenu::removeEntriesOnCurrentPage()
{
    foreach(QAction *action, m_actionGroup->actions())
    {
        if(m_pages.value(action) == m_currentPage)
        {
            m_actionGroup->removeAction(action);
            this->removeAction(action);

            m_pages.remove(action);
            action->deleteLater();
        }
    }
}

void BookmarksMenu::goToPreviousEntry()
{
    QAction *previous = 0;

    foreach(QAction *action, m_actionGroup->actions())
    {
        if(QPair<int, int>(m_pages.value(action), action->data().toInt()) < QPair<int, int>(m_currentPage, m_value))
        {
            if(previous)
            {
                previous = QPair<int, int>(m_pages.value(action), action->data().toInt()) > QPair<int, int>(m_pages.value(previous), previous->data().toInt()) ? action : previous;
            }
            else
            {
                previous = action;
            }
        }
    }

    if(previous)
    {
        previous->trigger();
    }
    else if(!m_actionGroup->actions().isEmpty())
    {
        QAction *last = m_actionGroup->actions().last();

        foreach(QAction *action, m_actionGroup->actions())
        {
            if(QPair<int, int>(m_pages.value(action), action->data().toInt()) > QPair<int, int>(m_pages.value(last), last->data().toInt()))
            {
                last = action;
            }
        }

        last->trigger();
    }
}

void BookmarksMenu::goToNextEntry()
{
    QAction *next = 0;

    foreach(QAction *action, m_actionGroup->actions())
    {
        if(QPair<int, int>(m_pages.value(action), action->data().toInt()) > QPair<int, int>(m_currentPage, m_value))
        {
            if(next)
            {
                next = QPair<int, int>(m_pages.value(action), action->data().toInt()) < QPair<int, int>(m_pages.value(next), next->data().toInt()) ? action : next;
            }
            else
            {
                next = action;
            }
        }
    }

    if(next)
    {
        next->trigger();
    }
    else if(!m_actionGroup->actions().isEmpty())
    {
        QAction *first = m_actionGroup->actions().first();

        foreach(QAction *action, m_actionGroup->actions())
        {
            if(QPair<int, int>(m_pages.value(action), action->data().toInt()) < QPair<int,int>(m_pages.value(first), first->data().toInt()))
            {
                first = action;
            }
        }

        first->trigger();
    }
}

void BookmarksMenu::clearList()
{
    foreach(QAction *action, m_actionGroup->actions())
    {
        m_actionGroup->removeAction(action);
        this->removeAction(action);

        m_pages.remove(action);
        action->deleteLater();
    }
}

void BookmarksMenu::setCurrentPage(int currentPage)
{
    m_currentPage = currentPage;
}

void BookmarksMenu::setValue(int value)
{
    m_value = value;
}

void BookmarksMenu::slotActionGroupTriggered(QAction *action)
{
    int page = m_pages.value(action);
    int value = action->data().toInt();

    emit entrySelected(page, value);
}

// settings dialog

SettingsDialog::SettingsDialog(QWidget *parent) : QDialog(parent)
{

    m_automaticRefreshCheckBox = new QCheckBox(this);
    m_openUrlCheckBox = new QCheckBox(this);

    m_antialiasingCheckBox = new QCheckBox(this);
    m_textAntialiasingCheckBox = new QCheckBox(this);
    m_textHintingCheckBox = new QCheckBox(this);

    m_maximumPageCacheSizeComboBox = new QComboBox(this);
    m_maximumPageCacheSizeComboBox->addItem(tr("%1 MB").arg(32), 33554432u);
    m_maximumPageCacheSizeComboBox->addItem(tr("%1 MB").arg(64), 67108864u);
    m_maximumPageCacheSizeComboBox->addItem(tr("%1 MB").arg(128), 134217728u);
    m_maximumPageCacheSizeComboBox->addItem(tr("%1 MB").arg(256), 268435456u);
    m_maximumPageCacheSizeComboBox->addItem(tr("%1 MB").arg(512), 536870912u);
    m_maximumPageCacheSizeComboBox->addItem(tr("%1 MB").arg(1024), 1073741824u);
    m_maximumPageCacheSizeComboBox->addItem(tr("%1 MB").arg(2048), 2147483648u);

    m_prefetchDistanceSpinBox = new QSpinBox(this);
    m_prefetchDistanceSpinBox->setRange(0, 10);

    m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel, Qt::Horizontal, this);
    connect(m_buttonBox, SIGNAL(accepted()), this, SLOT(accept()));
    connect(m_buttonBox, SIGNAL(rejected()), this, SLOT(reject()));

    m_layout = new QFormLayout(this);
    this->setLayout(m_layout);

    m_layout->addRow(tr("Automatic &refresh:"), m_automaticRefreshCheckBox);
    m_layout->addRow(tr("Open &URL:"), m_openUrlCheckBox);
    m_layout->addRow(tr("&Antialiasing:"), m_antialiasingCheckBox);
    m_layout->addRow(tr("&Text antialiasing:"), m_textAntialiasingCheckBox);
    m_layout->addRow(tr("Text &hinting:"), m_textHintingCheckBox);
    m_layout->addRow(tr("Maximum page cache &size:"), m_maximumPageCacheSizeComboBox);
    m_layout->addRow(tr("&Prefetch distance:"), m_prefetchDistanceSpinBox);
    m_layout->addRow(m_buttonBox);

    m_automaticRefreshCheckBox->setChecked(Document::automaticRefresh());
    m_openUrlCheckBox->setChecked(Document::openUrl());

    m_antialiasingCheckBox->setChecked(Document::antialiasing());
    m_textAntialiasingCheckBox->setChecked(Document::textAntialiasing());
    m_textHintingCheckBox->setChecked(Document::textHinting());

    m_prefetchDistanceSpinBox->setValue(Document::prefetchDistance());

    // maximumPageCacheSize

    bool addItem = true;

    for(int index = 0; index < m_maximumPageCacheSizeComboBox->count(); index++)
    {
        uint itemData = m_maximumPageCacheSizeComboBox->itemData(index).toUInt();

        if(itemData == Document::maximumPageCacheSize())
        {
            addItem = false;

            m_maximumPageCacheSizeComboBox->setCurrentIndex(index);
        }
    }

    if(addItem)
    {
        m_maximumPageCacheSizeComboBox->addItem(tr("%1 MB").arg(Document::maximumPageCacheSize() / 1024 / 1024), Document::maximumPageCacheSize());

        m_maximumPageCacheSizeComboBox->setCurrentIndex(m_maximumPageCacheSizeComboBox->count()-1);
    }
}

void SettingsDialog::accept()
{
    Document::setAutomaticRefresh(m_automaticRefreshCheckBox->isChecked());
    Document::setOpenUrl(m_openUrlCheckBox->isChecked());

    Document::setAntialiasing(m_antialiasingCheckBox->isChecked());
    Document::setTextAntialiasing(m_textAntialiasingCheckBox->isChecked());
    Document::setTextHinting(m_textHintingCheckBox->isChecked());

    Document::setMaximumPageCacheSize(m_maximumPageCacheSizeComboBox->itemData(m_maximumPageCacheSizeComboBox->currentIndex()).toUInt());
    Document::setPrefetchDistance(m_prefetchDistanceSpinBox->value());

    QDialog::accept();
}

// help dialog

HelpDialog::HelpDialog(QWidget *parent) : QDialog(parent)
{
    m_textBrowser = new QTextBrowser(this);
#ifdef DATA_INSTALL_PATH
    m_textBrowser->setSource(QUrl(QString(DATA_INSTALL_PATH) + "/help.html"));
#else
    m_textBrowser->setSource(QUrl("qrc:/miscellaneous/help.html"));
#endif

    m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Close, Qt::Horizontal, this);
    connect(m_buttonBox, SIGNAL(accepted()), this, SLOT(accept()));
    connect(m_buttonBox, SIGNAL(rejected()), this, SLOT(reject()));

    this->setLayout(new QVBoxLayout());
    this->layout()->addWidget(m_textBrowser);
    this->layout()->addWidget(m_buttonBox);
}

// unique application

UniqueApplication::UniqueApplication(const QString &name, QObject *parent) : QObject(parent),
    m_name(name),
    m_running(false),
    m_sharedMemory(0),
    m_localServer(0)
{
    m_sharedMemory = new QSharedMemory(name, this);

    if(m_sharedMemory->attach())
    {
        m_running = true;
    }
    else
    {
        m_running = false;

        if(!m_sharedMemory->create(1))
        {
            qDebug() << m_sharedMemory->errorString();
            return;
        }

        m_localServer = new QLocalServer(this);
        connect(m_localServer, SIGNAL(newConnection()), this, SLOT(slotNewConnection()));

        QLocalServer::removeServer("unique_application_" + name);

        if(!m_localServer->listen("unique_application_" + name))
        {
            qDebug() << m_localServer->errorString();
            return;
        }
    }
}

bool UniqueApplication::sendMessage(const QString &message)
{
    if(!m_running)
    {
        return false;
    }

    QLocalSocket localSocket;

    localSocket.connectToServer("unique_application_" + m_name, QIODevice::WriteOnly);
    if(!localSocket.waitForConnected())
    {
        qDebug() << localSocket.errorString();
        return false;
    }

    localSocket.write(message.toUtf8());
    if(!localSocket.waitForBytesWritten())
    {
        qDebug() << localSocket.errorString();
        return false;
    }

    localSocket.disconnectFromServer();

    return true;
}

void UniqueApplication::slotNewConnection()
{
    QLocalSocket *localSocket = m_localServer->nextPendingConnection();
    if(!localSocket->waitForReadyRead())
    {
        qDebug() << localSocket->errorString();
        return;
    }

    QString message = QString::fromUtf8(localSocket->readAll());

    localSocket->disconnectFromServer();

    emit messageReceived(message);
}
