/*

Copyright 2012 Adam Reichold

This file is part of qpdfview.

qpdfview is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

qpdfview is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with qpdfview.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MISCELLANEOUS_H
#define MISCELLANEOUS_H

#include <QtCore>
#include <QtNetwork>
#include <QtGui>

struct Link;
struct TocNode;
class Document;
class DocumentView;

// auxiliary view

class AuxiliaryView : public QWidget
{
    Q_OBJECT

public:
    explicit AuxiliaryView(QWidget *parent = 0);

    DocumentView *mainView() const;
    void setMainView(DocumentView *mainView);

protected:
    void showEvent(QShowEvent *event);

protected slots:
    virtual void slotDocumentChanged() = 0;

private:
    DocumentView *m_mainView;

};

// outline view

class OutlineView : public AuxiliaryView
{
    Q_OBJECT

public:
    explicit OutlineView(QWidget *parent = 0);

protected slots:
    void slotDocumentChanged();

    void slotItemClicked(QTreeWidgetItem *item, int column);

private:
    QTreeWidget *m_treeWidget;

    void prepareOutline(TocNode *node, QTreeWidgetItem *parent, QTreeWidgetItem *sibling);

};

// thumbnails view

class ThumbnailsView : public AuxiliaryView
{
    Q_OBJECT

    class PageItem : public QGraphicsItem
    {
    public:
        explicit PageItem(Document *document, int index, QGraphicsItem *parent = 0, QGraphicsScene *scene = 0);
        ~PageItem();

        QRectF boundingRect() const;
        void paint(QPainter *painter, const QStyleOptionGraphicsItem*, QWidget*);

    protected:
        void mousePressEvent(QGraphicsSceneMouseEvent*);

    private:
        Document *m_document;

        int m_index;

        QSizeF m_size;

        QFuture<void> m_render;
        void render();
    };

public:
    explicit ThumbnailsView(QWidget *parent = 0);

protected:
    void resizeEvent(QResizeEvent *event);

protected slots:
    void slotDocumentChanged();

private:
    QGraphicsScene *m_scene;
    QGraphicsView *m_view;

};

// presentation view

class PresentationView : public AuxiliaryView
{
    Q_OBJECT

public:
    explicit PresentationView();
    ~PresentationView();

public slots:
    void setCurrentPage(int currentPage);

    void previousPage();
    void nextPage();
    void firstPage();
    void lastPage();

protected:
    void paintEvent(QPaintEvent *event);
    void resizeEvent(QResizeEvent*);
    void keyPressEvent(QKeyEvent *event);
    void mousePressEvent(QMouseEvent *event);
    void mouseMoveEvent(QMouseEvent *event);

protected slots:
    void slotDocumentChanged();

private:
    int m_currentPage;

    QSizeF m_size;
    qreal m_scaleFactor;
    QRectF m_boundingRect;

    QList<Link> m_links;
    QTransform m_linkTransform;

    // internal methods

    void prepareView();

    QFuture<void> m_render;
    void render();
};

// recently used action

class RecentlyUsedAction : public QAction
{
    Q_OBJECT

public:
    RecentlyUsedAction(QObject *parent = 0);
    ~RecentlyUsedAction();

public slots:
    void addEntry(const QString &filePath);

    void clearList();

signals:
    void entrySelected(QString filePath);

protected slots:
    void slotActionGroupTriggered(QAction *action);

private:
    QMenu *m_menu;
    QActionGroup *m_actionGroup;
    QAction *m_separator;

    QAction *m_clearListAction;

    // settings

    QSettings m_settings;
};

// bookmarks menu

class BookmarksMenu : public QMenu
{
    Q_OBJECT

public:
    BookmarksMenu(QWidget *parent = 0);

public slots:
    void addEntry();
    void removeEntriesOnCurrentPage();

    void goToPreviousEntry();
    void goToNextEntry();

    void clearList();

    void setCurrentPage(int currentPage);
    void setValue(int value);

signals:
    void entrySelected(int page, int value);

protected slots:
    void slotActionGroupTriggered(QAction *action);

private:
    QActionGroup *m_actionGroup;

    QAction *m_addEntryAction;
    QAction *m_removeEntriesOnCurrentPageAction;

    QAction *m_goToPreviousEntryAction;
    QAction *m_goToNextEntryAction;

    QAction *m_clearListAction;

    QHash<QAction*, int> m_pages;

    int m_currentPage;
    int m_value;

};

// settings dialog

class SettingsDialog : public QDialog
{
    Q_OBJECT

public:
    explicit SettingsDialog(QWidget *parent = 0);

public slots:
    void accept();

private:
    QFormLayout *m_layout;
    QDialogButtonBox *m_buttonBox;

    QCheckBox *m_automaticRefreshCheckBox;
    QCheckBox *m_openUrlCheckBox;

    QCheckBox *m_antialiasingCheckBox;
    QCheckBox *m_textAntialiasingCheckBox;
    QCheckBox *m_textHintingCheckBox;

    QComboBox *m_maximumPageCacheSizeComboBox;

    QSpinBox *m_prefetchDistanceSpinBox;

};

// help dialog

class HelpDialog : public QDialog
{
    Q_OBJECT

public:
    explicit HelpDialog(QWidget *parent = 0);

    QSize sizeHint() const
    {
        return QSize(500, 700);
    }

private:
    QTextBrowser *m_textBrowser;
    QDialogButtonBox *m_buttonBox;

};

// unique application

class UniqueApplication : public QObject
{
    Q_OBJECT

public:
    UniqueApplication(const QString &name, QObject *parent = 0);

    bool isRunning() const
    {
        return m_running;
    }

public slots:
    bool sendMessage(const QString &message);

signals:
    void messageReceived(const QString &message);

protected slots:
    void slotNewConnection();

private:
    QString m_name;
    bool m_running;

    QSharedMemory *m_sharedMemory;
    QLocalServer *m_localServer;

};

#endif // MISCELLANEOUS_H
